/**
  Finite Element Method Wood Simulator application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2017-02-17 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <GL/glut.h>
#include <iostream>
#include <vector>
#include <math.h>
#include "svg_parse.h"
#include "osl/vec4.h" /*<- handy little vector class */
#include "osl/bbox2d.h"
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z


/***************** FEM Parameters *************/
float dragCoeff=100.0; /* artifical dissipation, lbf / ft/s */
float dt=1.0e-6; /* physics timestep, in seconds */
int subSteps=1000; /* physics timesteps per redraw */
float pix2fem=1.0/90; /* 90dpi seems to be standard Inkscape SVG file resolution. */
float fem2disc=12.0/1.0; /* fem units are feet; discretization is by inch */
float dscl=1.0e1; /* displacement magnification */
float fscl=1.0e5; /* force magnification (color/lbf) */

const char *model_name="test.svg";


/*********************** OpenGL Vertex Buffer Code **********************/

/* Class representing an OpenGL vertex */
class myVertex {
public:
	vec3 loc; /* 3D location of vertex */
	vec3 color; /* color of vertex */
	myVertex(vec3 l,vec3 n) :loc(l), color(n) {}
};

/* Class representing an OpenGL vertex buffer object */
class myMesh {
public:
	GLuint vb; /* OpenGL "handle" to a vertex buffer object */
	int vlen; /* Number of elements in our vertex buffer object */
	
	GLuint eb; /* Handle to an element buffer object */
	int elen;
	
	/* Set up an on-graphics-card buffer pointing to this vertex and index data */
	myMesh(const myVertex *v,int vN,const int *idx,int iN,GLenum mode=GL_STATIC_DRAW_ARB);
	
	/* Draw the vertices of this mesh */
	void draw(GLenum mode=GL_TRIANGLES,bool withColor=true);
	
	/* Destroy the graphics card's copy of our data */
	~myMesh();
};

/* Return the distance, in bytes, between these two pointers */
size_t ptrdiff(const void *a,const void *b) {
	return ((const char *)b)-((const char *)a);
}

/* Set up an on-graphics-card buffer pointing to this vertex data */
myMesh::myMesh(const myVertex *v,int vN,const int *idx,int iN,GLenum mode)
	:vb(0), vlen(vN), eb(0), elen(iN)
{
	glGenBuffersARB(1,&vb); 
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB,
		sizeof(v[0])*vlen,&v[0],mode);
	glVertexPointer(3,GL_FLOAT,sizeof(v[0]),
		(void *)ptrdiff(v,&v[0].loc));
	glColorPointer(3,GL_FLOAT,sizeof(v[0]),
		(void *)ptrdiff(v,&v[0].color));
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
	
	glGenBuffersARB(1,&eb); 
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,eb);
	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB,
		sizeof(idx[0])*elen,&idx[0],mode);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
}

/* Draw the vertices of this mesh */
void myMesh::draw(GLenum mode,bool withColor)
{
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,eb);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	if (withColor)
		glEnableClientState(GL_COLOR_ARRAY);
	else
		glDisableClientState(GL_COLOR_ARRAY);

	if (elen==0)
		glDrawArrays(mode,0,vlen); // just draw all vertices
	else
		glDrawElements(mode,elen,GL_UNSIGNED_INT,0); // use indices

	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);        
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
}

/* Destroy the graphics card's copy of our data */
myMesh::~myMesh()
{
	glDeleteBuffersARB(1,&vb);
	glDeleteBuffersARB(1,&eb);
}


/************** Pure FEM Code **********************/
/*
Simple physics routines
Orion Lawlor, olawlor@acm.org, 4/25/2001
*/

class FEM_Computation;

/*
A node is a single point; but represents a small area
of material.  It has a location, velocity, and sums a
net force at each timestep.
*/
class Node
{
public:
	vec3 rest; //Initial undeformed location (ft)
	vec3 loc; //Displacement from initial location (ft)
	vec3 vel; //Velocity (ft/s)
	vec3 force; //Force temporary (pound-force)
	vec3 strain; // strains on adjacent elements (for display only)
	float mass; //Node's mass (slug)
	float i_mass; //Inverse of node's mass (1/slug)
	
	Node(vec3 Nloc,float Nmass) {
		rest=Nloc;
		vel=loc=vec3(0);
		mass=Nmass;
		i_mass=1.0/Nmass;
	}
	
	//These form the timeloop-- clear, add forces, check constraints, update
	void clearForce() {
		force=strain=vec3(0.0);
		force.y+=mass*32.14; /* gravitational force */
	}
	void addForce(vec3 newForce) {force+=newForce;}
	void subForce(vec3 newForce) {force-=newForce;}
	
	void update(float dt) {
		vec3 acc=force*i_mass;
		loc=loc+vel*dt+acc*dt*dt*0.5;
		vel += acc*dt;
	}
};


/*
An element is a small chunk of matter.  It's 
bounded by nodes, from which it determines its
own deformation, and to which it applies forces.

This is a 2D constant strain triangle.
*/
class Element
{
public:
	class Material
	{
	public:
		float Y;//Young's modulus (lbf/sq foot, typ. 1-30 * 10^6)
		float v;//Poisson's ratio (pure, typ. 0.2-0.5)
		float axial;//Strain to stress, along coordinate axes
		float cross;//Strain to stress, between coordinate axes
		float shear;//Strain to stress, shear
		
		Material(float young,float poisson) 
		{
			Y=young;
			v=poisson;
			//Isotropic linear elastic strain/stress
			// (Przemieniecki 1968)
			float scale=Y/(1-v*v);
			axial=scale;
			cross=scale*v;
			shear=scale*(1-v)/2;
		}
	};
	class StrainState
	{
	public:
		float x,y,xy;//Strain on each axis
		//static float scale=10;
		float pin(float val) {
			if (val<0.0) return 0.0;
			if (val>1.0) return 1.0;
			return val;
		}
		/*
		float getX() {return pin(0.5+scale*x);}
		float getY() {return pin(0.5+scale*y);}
		float getXY() {return pin(0.5+scale*xy);}
		*/
	};
	
	int na,nb,nc; //Our three node numbers
	float xA,xB,xC;//Initial stress coefficients (x)
	float yA,yB,yC;//Initial stress coefficients (y)
	float areaInv; // 1.0/area
	bool fail; /* if true, this element is broken into pieces */
	//StrainState strain; // output strain values
	
	Element(float thick,int Na,int Nb,int Nc,FEM_Computation *fem);
	void step(FEM_Computation *fem);
};

/**
 Controls the overall computation--keeps the big lists of mesh entities,
 and calls them to get work done.
*/
class FEM_Computation 
{
public:
	// Mesh
	std::vector<Node> nodes;
	std::vector<Element> elements;
	
	// Boundary conditions
	class GluePair { public: int a,b; GluePair(int a_,int b_) :a(a_), b(b_) {} };
	std::vector<GluePair> glue_nodes; /* nodes that are pasted together */
	
	std::vector<int> fixed_nodes; /* nodes that shouldn't move */
	
	std::vector<int> load_nodes; /* nodes with an applied load force */
	std::vector<vec3> loads; /* corresponding loads to apply to above nodes */
	
	//Selection control
	int selectedNode;
	vec3 selPos;
	
	Element::Material *mat1;
	
	FEM_Computation(void) {
		selectedNode=-1;
	
		//Initialize the elements
		float young=2000000.00;//Young's modulus (lbs/sq inch), value for quality woods
		young*=12*12; /* convert to pounds per square foot */
		float poisson=0.4; //Poisson's ratio: 0.5 is incompressible in 2D
		mat1=new Element::Material(young,poisson);
		
		// Caller must insert nodes and elements
	}
	
	//Perform one physics step
	void physicsStep(float dt);
};

//Element creation: takes a set of nodes, determines rest configuration
Element::Element(float thickness,int Na,int Nb,int Nc,FEM_Computation *fem)
{
	na=Na; nb=Nb; nc=Nc;
	Node &a=fem->nodes[na];
	Node &b=fem->nodes[nb];
	Node &c=fem->nodes[nc];
	
	vec3 a0=a.rest, b0=b.rest, c0=c.rest;
	float x1=a0.x, y1=a0.y;
	float x2=b0.x, y2=b0.y;
	float x3=c0.x, y3=c0.y;
	xA=y2-y3; xB=y3-y1; xC=y1-y2;
	yA=x2-x3; yB=x3-x1; yC=x1-x2;
	
	vec3 B=b.rest-a.rest, C=c.rest-a.rest;
	areaInv=0.25/(B.y*C.x - B.x*C.y);
	if (areaInv<0) {
		printf("Warning: inside-out triangle on nodes %d,%d,%d\n",na,nb,nc);
		areaInv=-areaInv;
	}
	areaInv*=thickness; /* scales our forces by our thickness */
	fail=false;
	
}

/*Perform one element physics step: compute and apply node
forces based on internal strains.

  Equations from pp. 34-41 of 
  "The Finite Element Method for Engineers"
  by K. H. Huebner (Wiley & Sons 1975)

This version is *not* rotation independent 
(small strain linear version), so element will
change scale if they get rotated.
 */
void Element::step(FEM_Computation *fem) 
{
	Node &a=fem->nodes[na];
	Node &b=fem->nodes[nb];
	Node &c=fem->nodes[nc];
	if (fail) {
		vec3 fail_strain(999.0f,999.0f,999.0f); /* show broken wood as white */
		a.strain+=fail_strain;
		b.strain+=fail_strain;
		c.strain+=fail_strain;
		return; /* no forces from broken elements */
	}

	vec3 B=b.loc-a.loc;
	vec3 C=c.loc-a.loc;

	//Compute strains based on simple linear model:
	float strain_x=+areaInv*(B.x*xB+C.x*xC);
	float strain_y=-areaInv*(B.y*yB+C.y*yC);
	float strain_xy=areaInv*(B.y*xB-B.x*yB + C.y*xC-C.x*yC);
//	System.out.println("sx="+strain_x+" sy="+strain_y+" sxy="+strain_xy);
	
	// Output our strain to adjacent nodes (for visualization)
	float linear=strain_x+strain_y;
	float shear=fabs(strain_xy);
	vec3 strain(linear>0?linear:0, linear>0?0:-linear, shear);
	a.strain=b.strain=c.strain=strain;

	//Compute stresses from strains
	const Material &mat=*fem->mat1;
	float stress_x=+mat.axial*strain_x+mat.cross*strain_y;
	float stress_y=+mat.cross*strain_x+mat.axial*strain_y;
	float stress_xy=mat.shear*strain_xy;
	
	/* Breaking stress, in pounds per square foot */
	const float max_stress=350*12*12;
 /* wood basically only fails in tension or shear */
	if ((stress_x+stress_y)<-max_stress || fabs(stress_xy)>max_stress)
	{ /* woa--this element just failed! */
		fail=true;
	}
	
	/**
	  Element viscosity/drag coefficient (N/(m/s)) 
	  Drag tend to push node velocities toward local average
	  This allows a finite timestep, by draining away short waves.
	*/	
	extern float dragCoeff;
	float d=dragCoeff;
	vec3 avg=(a.vel+b.vel+c.vel)*(1.0/3);

	//Stresses apply forces to nodes
	float s=-0.5;
	a.addForce(vec3(
		s*(+xA*stress_x-yA*stress_xy)+d*(avg.x-a.vel.x),
		s*(-yA*stress_y+xA*stress_xy)+d*(avg.y-a.vel.y),0));
	b.addForce(vec3(
		s*(+xB*stress_x-yB*stress_xy)+d*(avg.x-b.vel.x),
		s*(-yB*stress_y+xB*stress_xy)+d*(avg.y-b.vel.y),0));
	c.addForce(vec3(
		s*(+xC*stress_x-yC*stress_xy)+d*(avg.x-c.vel.x),
		s*(-yC*stress_y+xC*stress_xy)+d*(avg.y-c.vel.y),0));
}

//Perform one physics step
void FEM_Computation::physicsStep(float dt)
{
	unsigned int n;
	int e,el; el=elements.size();
	//Clear the node forces
	for (n=0;n<nodes.size();n++) nodes[n].clearForce();

	//Each element applies forces to its nodes
//#pragma omp parallel for schedule(static,1000)
	for (e=0;e<el;e++)
		elements[e].step(this);

	//Apply boundary conditions (external loads)
	for (n=0;n<load_nodes.size();n++)
		nodes[load_nodes[n]].force+=loads[n];

	//Update node positions based on forces
	for (n=0;n<nodes.size();n++)
		nodes[n].update(dt);
	
	//Lock down boundary nodes
	for (n=0;n<fixed_nodes.size();n++) {
		Node &a=nodes[fixed_nodes[n]];
		a.loc=a.vel=vec3(0.0);
	}
	for (n=0;n<glue_nodes.size();n++) {
		Node &a=nodes[glue_nodes[n].a];
		Node &b=nodes[glue_nodes[n].b];
		// Displacements on a glued pair apply equally to both nodes.
		// This keeps the two nodes locked together.
		// FIXME: it'd be nice if a glued joint could rotate through large angles...
		a.loc=b.loc=0.5*(a.loc+b.loc);
		a.vel=b.vel=0.5*(a.vel+b.vel);
	}
	
	if (selectedNode!=-1) 
	{ /* Try to *gently* impose a displacement */
		vec3 sel=nodes[selectedNode].rest; /* center of disp. */
		vec3 nu=selPos-sel; /* displacement to apply */
		float w=dt*1.0e2; /* displacement intensity */
		for (n=0;n<nodes.size();n++) {
			vec3 d=nodes[n].rest-sel;
			const float sel_len=0.5; /* FEM-units selection radius */
			const float sel_ilen2=(1.0/(sel_len*sel_len));
			float len2=dot(d,d); /* length(d) squared */
			float weight=w*(1.0-sel_ilen2*len2); /* fraction/second */
			if (weight>0) {
				nodes[n].loc=nodes[n].loc+weight*(nu-nodes[n].loc);
			}
		}
	}
}
	

/******** Rendering and integration code *********/
vec3 makeVertex(double x,double y) {
        return vec3(x,y,0); 
}
vec3 makeNormal(double x,double y) {
	return vec3(0,0,0); 
/*
        vec3 c=makeVertex(x,y);
        vec3 a=makeVertex(x+0.0001,y);
        vec3 b=makeVertex(x,y+0.0001);
        return (a-c).cross(b-c).dir();
*/
}

vec2 makeVec2(const vec3 &v) { return vec2(v.x,v.y); }
/* Return true if this vector is in the unit square */
bool inbounds(const vec2 &v) {
	float tol=0.001; /* enlargement to apply */
	if (v.x<-tol || v.x>1+tol) return false;
	if (v.y<-tol || v.y>1+tol) return false;
	return true;
}

// Return the area of the image of the unit square under this matrix.
double area(const mat2 &m) {
	return fabs(m(0,0)*m(1,1)-m(1,0)*m(0,1));
}

/* FEM Interface code--everything plugs together here. */
class FEM_Driver : public svg_consumer {
public:
	myMesh *m; /* GPU vertex buffer object */
	FEM_Computation fem;	
	std::vector<myVertex> vertices; // OpenGL vertex array, before GPU upload
	std::vector<int> indices; // triangle strip indices, for rendering 
	
	std::vector<mat2> fixed; // fixed supports
	std::vector<mat2> load; // applied loads
	std::vector<mat2> glue; // glued regions
	
	FEM_Driver();
	void draw(void);
	~FEM_Driver() {delete m; }
	
	// Mouse grabs and drags a portion of the mesh
	void mouse(int x,int y,int is_down);

private:
	osl::Bbox2d box;
	// Add a 2D mesh with this resolution, across the image of the unit square on this matrix.
	void addMesh(int w,int h,double thickness,const mat2 &m);
	
	virtual void svg_use(const char *what,const mat2 &where);
        virtual void svg_path(const std::vector<vec2> &pts,const svg_style &style);
	
// Matrix utilities used during boundary condition application
	mat2 curMatrix;
	// We're going to use this matrix's unit image for something;
	//  return the inverse.
	void setMatrix(const mat2 &m) {
		box.add(m.apply(vec2(0,0)));
		box.add(m.apply(vec2(1,1)));
		curMatrix=inverse(m);
	}
	// Return true if this node sits inside the current matrix
	bool nodeInMatrix(int nodeNo) {
		vec2 v=curMatrix.apply(makeVec2(fem.nodes[nodeNo].rest));
		return inbounds(v);
	}

// Drawing utilities
	vec2 center, scale;
	// Return the display location of this node
	vec3 nodeLoc(int n) const;
};


FEM_Driver::FEM_Driver()
{
	if (getenv("DT")) {dt=atof(getenv("DT"));}
	if (getenv("DRAG")) {dragCoeff=atof(getenv("DRAG"));}
	printf("Running simulation with dt=%.6g and dragCoeff=%.6g\n",dt,dragCoeff);
	
	/* Read SVG elements from file */
	box.empty();
	::svg_parse(model_name,*this);
	
	/* Upload mesh to GPU */
	int len=vertices.size();
	printf("Created %d vertices and %d indices\n",len,(int)indices.size());
	m=new myMesh(&vertices[0],len,&indices[0],indices.size(),GL_STREAM_DRAW_ARB); // 0,0); // 
	vertices.erase(vertices.begin(),vertices.end()); 
	indices.erase(indices.begin(),indices.end());
	
	/* Apply boundary conditions */
	unsigned int nGlue=0, nSkip=0;
	for (unsigned int g=0;g<glue.size();g++) {
		setMatrix(glue[g]);
		std::vector<int> glueNodes;
		
		// Find all nodes touched by the glue
		for (unsigned int n=0;n<fem.nodes.size();n++) {
			if (nodeInMatrix(n))
				glueNodes.push_back(n);
		}
		
		// Pair off the glue nodes
		unsigned int sg=0;// start glue
		while (sg+1<glueNodes.size()) // we still have >=2 nodes to glue
		{
			vec3 sL=fem.nodes[glueNodes[sg]].rest;
			int eg=-1; // end glue
			float ed=0.8/fem2disc; /* maximum search distance is a fraction of a typical discretized element */
			/* Find a nearby node to glue this one to */
			for (unsigned int tg=sg+1;tg<glueNodes.size();tg++) {
				vec3 tL=fem.nodes[glueNodes[tg]].rest;
				float td=length(sL-tL);
				if (td<ed) { /* New closest nearby node */
					eg=tg;
					ed=td;
				}
			}
			if (eg!=-1) 
			{ /* Glue sg to nearby node eg */
				fem.glue_nodes.push_back(FEM_Computation::GluePair(
					glueNodes[sg],glueNodes[eg]
				));
				sg++;
				nGlue++;
				
				/* Swap eg to end, and remove from glueNodes array */
				int last=glueNodes.size()-1;
				std::swap(glueNodes[eg],glueNodes[last]);
				glueNodes.erase(glueNodes.begin()+last);
			} else 
			{ /* sg is a bad node, with no neighbors--skip it */
				sg++;
				nSkip++;
			}
		}
	}
	printf("Applied glue to %d nodes (but had to skip %d)\n",nGlue, nSkip);
	
	for (unsigned int f=0;f<fixed.size();f++) {
		setMatrix(fixed[f]);
		for (unsigned int n=0;n<fem.nodes.size();n++) {
			if (nodeInMatrix(n))
				fem.fixed_nodes.push_back(n);
		}
	}
	printf("Found %d fixed nodes\n",(int)fem.fixed_nodes.size());
	
	for (unsigned int l=0;l<load.size();l++) {
		/* Figure out how much force to apply */
		float area2force=100.0; // lbs of force per square foot of load-symbol area
		if (getenv("LOAD_BIG")!=0) area2force=1000.0;
		double total=area(load[l])*area2force; // Total load (pounds)
		printf("Load %d: total of %.1f pounds of force\n",l,total);
		setMatrix(load[l]);
		
		/* Figure out how many nodes to apply our force to */
		unsigned int n,count=0;
		for (n=0;n<fem.nodes.size();n++) {
			if (nodeInMatrix(n)) count++;
		}
		double perNode=total/count;// Load per FEM node
		printf("Load %d: Applying total force of %.2f lbs to %d nodes\n",l,total,count);
		if (count==0) {
			printf("	WARNING!  Load %d doesn't seem to touch anything!  Ignoring the load!\n",l);
		}
		
		/* Apply force to those nodes */
		for (n=0;n<fem.nodes.size();n++) {
			if (nodeInMatrix(n))
			{
				fem.load_nodes.push_back(n);
				fem.loads.push_back(vec3(0,+1,0)*perNode);
			}
		}
	}
	printf("Applied forces at %d nodes\n",(int)fem.load_nodes.size());
}

void FEM_Driver::svg_use(const char *what,const mat2 &where)
{
	mat2 f=where; /* convert to FEM units */
	f.setO(f.apply(vec2(-1000,0))*pix2fem); /* all groups start at -1000 pixels */
	if (f(0,2)<0) return; /* ignore left-of-origin (labeling) uses */
	float s=50.0*pix2fem; /* load and support groups are 50 pixels across*/
	f.setX(f.getCol(0)*s);
	f.setY(f.getCol(1)*s);
	
	/* Subtle: we have to just stash these objects for now,
	because we might not have the whole mesh yet (depends on file order) 
	*/
	std::string w=what;
	if (w=="#Load") load.push_back(f);
	if (w=="#Support") fixed.push_back(f);
	if (w=="#Glue") glue.push_back(f);
}

void FEM_Driver::svg_path(const std::vector<vec2> &pts,const svg_style &style)
{
	if (2==pts.size()) 
	{ /* Line segment--generate a mesh here */
		vec2 a=pts[0]*pix2fem, b=pts[1]*pix2fem;
		if (a.x<0) return; /* sample geometry to left of origin is not simulated */
		vec2 x=(b-a);
		float length=x.mag();
		float width=style.width*pix2fem;
		float thickness=style.opacity;
		vec2 y=x.perp()*width/length;
		vec2 origin=a-0.5f*y;
		mat2 m(1.0); 
		m(0,0)=x.x; m(1,0)=x.y;
		m(0,1)=y.x; m(1,1)=y.y;
		m(0,2)=origin.x; m(1,2)=origin.y;
		int w=1+(int)(length*fem2disc+0.5);
		int h=1+(int)(width*fem2disc+0.5);
		addMesh(w,h,thickness,m);
	} else {
		/* else just ignore it */
	}
}

void FEM_Driver::addMesh(int w,int h,double thickness,const mat2 &m)
{
	double density=0.5; /* grams/cc; typical value for douglas fir */
	density*=1.94; /* convert from grams/cc to slugs/cubic foot (lovely US units)*/
	double nodeMass=area(m)/(w*h)*thickness*density; /* mass/node (slugs) */
	
	printf("Creating a %d x %d mesh\n",w,h);
	for (int iy=0;iy<h;iy++) {
		double fy=iy*(1.0/(h-1));
		for (int ix=0;ix<w;ix++) {
			double fx=ix*(1.0/(w-1));
			vec2 v=m.apply(osl::Vector2d(fx,fy));
			box.add(v);
			double x=v.x, y=v.y;
			int idx=vertices.size();
		
			/* make rendering vertex */
			vec3 loc=makeVertex(x,y);
			vertices.push_back(myVertex(loc,loc));
			if (iy<h-1) { /* make rendering triangle strip */
				if (ix==0) indices.push_back(idx);
				indices.push_back(idx);
				indices.push_back(idx+w);
				if (ix==w-1) indices.push_back(idx+w);
			}
		
			/* make FEM vertex */
			fem.nodes.push_back(Node(loc,nodeMass));
		
			/* make FEM elements */
			if (ix>0 && iy>0) { 
				fem.elements.push_back(Element(
					thickness,idx-1-w,idx-1,idx-w,&fem
				));
				fem.elements.push_back(Element(
					thickness,idx-w,idx-1,idx,&fem
				));
			}
		}
	}
}

vec3 FEM_Driver::nodeLoc(int n) const
{
	return fem.nodes[n].rest+fem.nodes[n].loc*dscl;
}

void FEM_Driver::draw(void)
{
	glutPostRedisplay();

/* Set up OpenGL to match FEM coordinate system */
	scale=vec2(1.95/box.getWidth());
	scale.y=-scale.y; /* to match inkscape y-is-down coordinates */
	scale.y*=glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT);
	glScalef(scale.x,scale.y,0);
	center=box.getCenter();
	glTranslatef(-center.x,-center.y,0);
	
/* Run physics for 20ms */
	int end_time=glutGet(GLUT_ELAPSED_TIME)+20;
	while (glutGet(GLUT_ELAPSED_TIME)<end_time) fem.physicsStep(dt);

/* Draw background grid */
	glLineWidth(1.0);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_DEPTH_TEST);
	glColor4f(0,0,0,0.1);
	glBegin(GL_LINES);
	int limit=100;
	for (int x=-limit;x<limit;x++)
		{glVertex2f(x,-limit); glVertex2f(x,+limit);}
	for (int y=-limit;y<limit;y++)
		{glVertex2f(-limit,y); glVertex2f(+limit,y);}
	glEnd();
	
/* Upload new node positions to GPU */
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,m->vb);
	myVertex *vd=(myVertex *)glMapBufferARB(GL_ARRAY_BUFFER_ARB,GL_WRITE_ONLY_ARB); /* vertex data pointer */
	float dmax=0.0;
	for (unsigned int i=0;i<fem.nodes.size();i++) {
		float d=length(fem.nodes[i].loc);
		dmax=std::max(dmax,d);
		vd[i].loc=nodeLoc(i);
		if (0) { // Show forces as node color
			vd[i].color=fem.nodes[i].force*fscl+vec3(0.5f);
			//fem.nodes[i].vel*50.0+vec3(0.5f);
			vd[i].color.z=(fem.nodes[i].force.x+fem.nodes[i].force.y)*fscl+0.5f;
		}
		if (1) { // Show stress as node color
      vec3 v=fem.nodes[i].strain*fscl;
			// wraparound: v=vec3(fmod(v.x,1.0),fmod(v.y,1.0),fmod(v.z,1.0));
      vd[i].color=v;
		}
	}
	glUnmapBufferARB(GL_ARRAY_BUFFER_ARB);
	m->draw(GL_TRIANGLE_STRIP);
	
	/* Draw mesh outline */
	glColor4f(0.5,0.5,0.5,0.3);
	glLineWidth(0.8);
	//m->draw(GL_LINES,false);
	glPointSize(0.8);
	m->draw(GL_POINTS,false);
	
	/* Draw mesh glue */
	glLineWidth(2.0);
	glColor4f(0.8,0.5,0.0,0.3);
	glBegin(GL_LINES);
	for (unsigned int g=0;g<fem.glue_nodes.size();g++) {
		glVertex2fv(nodeLoc(fem.glue_nodes[g].a));
		glVertex2fv(nodeLoc(fem.glue_nodes[g].b));
	}
	glEnd();

/* Draw fixed boundary conditions */
	glBegin(GL_QUADS);
	for (unsigned int f=0;f<fixed.size();f++) {
		glColor3f(0,0,0);
		glVertex2dv(fixed[f].apply(vec2(0,0)));
		glVertex2dv(fixed[f].apply(vec2(1,0)));
		glColor3f(1,1,1);
		glVertex2dv(fixed[f].apply(vec2(1,1)));
		glVertex2dv(fixed[f].apply(vec2(0,1)));
	}
	glEnd();
	
	static int last_time=end_time;
	if (end_time>last_time+200) { /* print displacement every 200ms */
		printf("Maximum displacement: %.2f ft / %.3f in\n",dmax,12.0*dmax);
		last_time=end_time;
	}
	
	if (dmax>100.0) { /* er, the mesh just exploded.  Cut the timestep and reset it. */
		dt*=0.5;
		printf("Instability detected.  Reducing timestep to %.3g s and resettting simulation.\n",dt);
		for (unsigned int n=0;n<fem.nodes.size();n++) {
			fem.nodes[n].loc=vec3(0.0);
			fem.nodes[n].vel=vec3(0.0);
		}
	}
}
// Mouse grabs and drags a portion of the mesh
void FEM_Driver::mouse(int x,int y,int is_down) {
	vec3 loc(
		2.0*x/float(glutGet(GLUT_WINDOW_WIDTH))-1.0,
		1.0f-2.0*y/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.0); /* to OpenGL coords */
	loc=vec3(loc.x/scale.x+center.x,loc.y/scale.y+center.y,0); /* to FEM coords */
	static int was_down=0;
	if (is_down && !was_down) { /* going down */
		int closestNode=0;
		float closest=1.0e10;
		for (unsigned int i=0;i<fem.nodes.size();i++)
		{
			float d=length(loc-nodeLoc(i));
			if (d<closest) {
				closest=d;
				closestNode=i;
			}
		}
		fem.selectedNode=closestNode;
	}
	if (!is_down) fem.selectedNode=-1;

	fem.selPos=loc;

	was_down=is_down;
}

FEM_Driver *hf;

void my_display(void) {
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */

	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	if (hf==0) hf=new FEM_Driver;
	hf->draw();
	
	glutSwapBuffers(); /* copy back buffer to front */

#if 0
/* Statistics */
	static int last_msec=0;
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
#endif
}

void my_mouse(int state,int buttons,int x,int y) {
	hf->mouse(x,y,state!=0);
}
void my_motion(int x,int y) { 
	hf->mouse(x,y,1);
}
void my_key(unsigned char key,int x,int y) {
	if (key=='q' || key=='x') exit(0);
}

int main(int argc,char *argv[]) {
	if (argc>1) model_name=argv[1];
	
	glutInitWindowSize(900,900);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	glutCreateWindow("OpenGL 2D Finite Element Wood Simulator");
	glewInit();
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	glutKeyboardFunc(my_key);
	
	glutMainLoop(); /* processes events */
	return 0;
}
